(in-package :smots)

(defgeneric language (specification)
  (:documentation "current language of SPECIFICATION"))

(defgeneric (setf language) (val specification)
 (:documentation "writer on current language of SPECIFICATION"))

(defgeneric languages (specification)
  (:documentation "list of defined languages in SPECIFICATION"))

(defgeneric (setf languages) (val specification)
 (:documentation "writer on the list of defined languages in SPECIFICATION"))

(defun make-languages () (make-hash-table :test #'equal))

(defclass language (alanguage)
  ((words :initform nil :accessor words :initarg :words)))

(defmethod initialize-instance :after
    ((language language) &rest initargs &key &allow-other-keys)
  (declare (ignore initargs))
  (let ((cardinal (compute-cardinal language)))
    (setf (slot-value language 'computed-cardinal) cardinal)))

(defun languagep (o)
  (typep o 'language))

(defmethod print-object ((language language) stream)
  (format stream "{")
  (display-sequence (words language) stream :sep ", ")
  (format stream "} "))

(defun make-language (words wordtype languagetype)
  (let ((words (remove-duplicates words :from-end t :test #'eq)))
    (if (and words (null (cdr words)))
	(car words)
	(make-instance 'language
		       :words words
		       :name "unmaned"
		       :wordtype wordtype
		       :languagetype languagetype))))

(defgeneric empty-language (alanguage)
  (:documentation "creates empty language according to wordtype"))

(defmethod empty-language ((alanguage alanguage))
  (make-instance 'language :name "empty"
		 :wordtype (wordtype alanguage)
		 :languagetype (languagetype alanguage)))

(defun empty-languagep (l)
  (and (languagep l) (endp (words l))))

(defmethod lintersection ((language1 language) (language2 language))
  (lgeneric-bool language1 language2 #'intersection))

(defmethod lunion ((language1 language) (language2 language))
  (lgeneric-bool language1 language2 #'union))

(defgeneric lset-difference (language1 language2)
  (:documentation "difference of LANGUAGE1 and LANGUAGE2"))

(defmethod lset-difference ((language1 language) (language2 language))
  (lgeneric-bool language1 language2 #'set-difference))

(defgeneric lconcatenation (language1 language2)
  (:documentation "concatenation of LANGUAGE1 and LANGUAGE2"))

(defmethod lconcatenation ((language1 language) (language2 language))
  (lgeneric language1 language2 #'concatenation))

(defmethod lmirror ((language language))
  (make-language
   (mapcar #'mirror (words language))
   (wordtype language)
   (languagetype language)))

(defgeneric mix (language1 language2)
  (:documentation "concatenation of LANGUAGE1 and LANGUAGE2"))

(defmethod mix :before (language1 language2)
  (assert (disjoint-alphabets (list language1 language2))))

(defmethod mix ((language1 language) (language2 language))
  (lgeneric language1 language2 #'mix))

(defmethod extreme-letters ((language language) &optional (last nil))
;;  (format *error-output* "extreme-letters language~%")
  (extreme-letters (words language) last))

(defmethod compute-cardinal ((language language))
  (length (words language)))

(defmethod equivalent ((language1 language) (language2 language))
  (let ((words1 (words language1))
	(words2 (words language2)))
    (and (subsetp words1 words2) (subsetp words2 words1))))

(defmethod computed ((language language))
  t)

(defmethod lgeneric-bool ((language1 language) (language2 language) boolfun)
  (make-language
   (words-generic-bool (words language1) (words language2) boolfun)
   (wordtype language1)
   (languagetype language1)))
  
(defmethod lgeneric ((language1 language) (language2 language) fun)
  (make-language
   (words-generic (words language1) (words language2) fun)
   (wordtype language1)
   (languagetype language1)))

(defmethod ljoin ((language1 language) (language2 language))
;;  (format *error-output* "ljoin language langage ~%")
  (lgeneric language1 language2 #'ljoin))
 
(defmethod compute-language ((language language))
;;  (format *error-output* "compute-language language~%")
  language)

(defmethod compute-prefix-language ((language language) (n integer))
;;  (format *error-output* "compute-prefix-language language~%")
  (make-language (remove-if (lambda (word) (> (wlength word) n))
			    (words language))
		 (wordtype language)
		 (languagetype language)))

(defmethod cut-extreme-letter ((lexprs list) (letter abstract-letter) &optional (last nil))
  (format *error-output* "cut-extreme-letter list~%")
  (mapcar (lambda (lexpr)
	    (cut-extreme-letter lexpr letter last))
	  lexprs))

(defmethod cut-extreme-letter ((language language) (letter abstract-letter) &optional (last nil))
  (format *error-output* "cut-extreme-letter language")
  (make-language
   (remove-if #'empty-languagep
	      (cut-extreme-letter (words language) letter last))
   (wordtype language)
   (languagetype language)))

(defmethod concatenation ((language1 language) (language2 language))
  (lgeneric language1 language2 #'concatenation))

(defmethod priority ((language language))
  6)

(defmethod weight ((language language))
  (declare (ignore language))
  1)

(defmethod projection ((language language) (alphabet alphabet))
  (let ((words (projection (words language) alphabet)))
    (make-language words (type-of language) (languagetype language))))

(defun add-language (language languages)
  (setf (gethash (name language) languages) language))

(defun add-current-language ()
  (add-language (language *spec*) (languages *spec*)))

(defun set-current-language (language)
  (setf (language *spec*) language)
  (add-current-language))

(defmethod cardinal ((language language))
  (length (words language)))

(defun write-language (language stream)
  (format stream "Language ~A~%" (name language))
  (unless (empty-wordp language)
    (format stream "~A~%" language)))
  
(defun write-current-language (stream)
  (write-language (language *spec*) stream))

(defun save-language (file)
  (back-up-file (absolute-filename file))
  (with-open-file (foo (absolute-filename file) :direction :output)
    (if foo
        (progn
          (write-current-language foo)
          (format *standard-output* "Language saved~%"))
        (prog1
            nil
          (format *error-output* "unable to open ~A~%" file)))))

(defmethod compute-prefix-language :around ((alanguage alanguage) (n integer))
  (format *error-output* "compute-prefix-language around~%")
  (if (zerop n)
      (if (contains-empty-sword alanguage)
	  (empty-sword)
	  (empty-language alanguage))
      (call-next-method)))

(defmethod gen-calphabet-of ((language language) lettertype)
;;  (format *error-output* "gen-alphabet-of language t")
  (gen-alphabet-of (words language) lettertype))
